from __future__ import annotations

import itertools
import logging
from os import PathLike
from pathlib import Path
from typing import Any, Callable, Optional, overload, Type

# import file_parser
from .file_parser import FileParser, _get_parser_from_path
from .joy_map import JoyMap
from .key_map import KeyBind, KeyMap
from .base_manager import BaseManager, _CallableSets

import pygame

logger: logging.Logger = logging.getLogger(__name__)


class KeyListener(BaseManager):
    _listeners: dict[str, KeyListener] = {}
    key_map: KeyMap = KeyMap()
    joy_map: JoyMap = JoyMap()

    def __init__(self, handle: str) -> None:
        super().__init__(handle)

        # --------Basic function assignment--------
        # Meta dict, string bind name as key, then event type as key to get callable
        # get(bind_name).get(concurrency).get(event_type)
        self._key_hooks: dict[str, dict[bool, dict[int, list[Callable]]]] = {}

        # --------Class method assignment--------
        # Meta dict, string bind name as key, then Pygame event key, method and
        # affected object as values
        # get(bind_name).get(concurrency).get(event_type)
        self._class_listeners: dict[
            str, dict[bool, dict[int, list[tuple[Callable, Type[object]]]]]
        ] = {}
        # Inversion of _class_listeners. Method as key, bind name
        self._class_listener_binds: dict[Callable, list[str]] = {}

    @overload
    def bind(
        self,
        key_bind_name: str,
        default_key: Optional[int],
        default_mod: Optional[int],
        event_type: Optional[int],
    ) -> Callable:
        """
        Adds a bind field to the key registry, and associates the following
        callable with that field so when the key associated with the field is
        pressed, the callable is called. If the field does not exist, it
        is created.

        :param key_bind_name: Reference name for the binding function.
        Callable events will be hooked by this name.
        :param default_key:
        Pygame key constant used to fill the registry if the bind doesn't
        exist or does not have an assigned key, defaults to None
        :param default_mod: Mod keys required for activating the key bind. None means
        the bind works with any mod keys pressed. pygame.KMOD_NONE means it requires no
        mod keys to be pressed. If using multiple, use bitwise OR to combine, defaults
        to None
        :param event_type: Event that the the bind is to be called on. Must be
        specified as a keyword. For keys, should be either pygame.KEYDOWN or
        pygame.KEYUP. Defaults to pygame.KEYDOWN
        """

    @overload
    def bind(
        self,
        key_bind_name: str,
        default_key: Optional[int] = None,
        default_mod: Optional[int] = None,
    ) -> Callable:
        """
        Adds a bind field to the key registry, and associates the following
        callable with that field so when the key associated with the field is
        pressed, the callable is called. If the field does not exist, it
        is created.

        :param key_bind_name: Reference name for the binding function.
        Callable events will be hooked by this name.
        :param default_key:
        Pygame key constant used to fill the registry if the bind doesn't
        exist or does not have an assigned key, defaults to None
        :param default_mod: Mod keys required for activating the key bind. None means
        the bind works with any mod keys pressed. pygame.KMOD_NONE means it requires no
        mod keys to be pressed. If using multiple, use bitwise OR to combine, defaults
        to None
        """

    @overload
    def bind(
        self,
        key_bind_name: str,
        default_joystick_data: Optional[dict],
        event_type: Optional[int],
    ) -> Callable:
        """
        Adds a bind field to the key registry, and associates the following
        callable with that field so when the key associated with the field is
        pressed, the callable is called. If the field does not exist, it
        is created.

        :param key_bind_name: Reference name for the binding function.
        Callable events will be hooked by this name.
        :param default_joystick_data: A dictionary containing the needed data for a the
        desired joystick event. See documentation for more information.
        :param event_type: Event that the the bind is to be called on. Must be
        specified as a keyword. For joystick/controller, should be a joystick event. See
        documentation for full list.
        """

    @overload
    def bind(
        self, key_bind_name: str, default_joystick_data: Optional[dict] = None
    ) -> Callable:
        """
        Adds a bind field to the key registry, and associates the following
        callable with that field so when the key associated with the field is
        pressed, the callable is called. If the field does not exist, it
        is created.

        If the event type is not supplied, it is intuited from the joystick data.

        :param key_bind_name: Reference name for the binding function.
        Callable events will be hooked by this name.
        :param default_joystick_data: A dictionary containing the needed data for a the
        desired joystick event. See documentation for more information.
        """

    def bind(self, key_bind_name: str, *args, **kwds) -> Callable:
        event_type: int
        is_stick = False
        default_key: Optional[int] = kwds.get("default_key", None)
        default_mod: Optional[int] = kwds.get("default_mod", None)
        default_joystick_data: Optional[dict] = kwds.get("default_joystick_data", None)
        for index, arg in enumerate(args):
            if type(arg) is dict:
                is_stick = True
                default_joystick_data = arg
                break
            if index == 0:
                default_key = arg
            elif index == 1:
                default_mod = arg

        if is_stick:
            self.joy_map.generate_bind(key_bind_name, default_joystick_data)
            # Attempt to intuit the desired event from the provided dict
            default_stick_event = pygame.JOYBUTTONDOWN
            if default_joystick_data is not None:
                joy_keys = default_joystick_data.keys()
                if "axis" in joy_keys:
                    default_stick_event = pygame.JOYAXISMOTION
                elif "hat" in joy_keys:
                    default_stick_event = pygame.JOYHATMOTION
            event_type = kwds.get("event_type", default_stick_event)
        else:
            self.key_map.generate_bind(key_bind_name, default_key, default_mod)
            event_type = kwds.get("event_type", pygame.KEYDOWN)

        def decorator(responder: Callable) -> Callable:
            # Regardless, add the responder to the bind within our hook dict
            is_concurrent = not hasattr(responder, "_runs_sequential")
            event_dict = self._key_hooks.setdefault(key_bind_name, {})
            hooks = event_dict.setdefault(is_concurrent, {})
            listeners = hooks.setdefault(event_type, [])
            if responder not in listeners:
                listeners.append(responder)
            return responder

        return decorator

    @overload
    def rebind(
        self,
        key_bind_name: str,
        new_key: Optional[int] = None,
        new_mod: Optional[int] = None,
    ) -> tuple[int | None, int | None] | None:
        """
        Attempts to assign the new key info the the named bind.
        Generates a warning if the bind is not registered.

        :param key_bind_name: Name of the bind to be reassigned
        :param default_key: Pygame key to be assigned to
        :param default_mod: Pygame mod keys for assignment,
        defaults to None
        :return: A tuple containing the previous key and mod key
        """

    @overload
    def rebind(
        self,
        key_bind_name: str,
        new_joystick_data: Optional[dict] = None,
    ) -> dict | None:
        """
        Attempts to assign the new joystick data to the named bind.
        Generates a warning if the bind is not registered.

        :param key_bind_name: Name of the bind to be reassigned
        :param new_joystick_data: Joystick data to that needs to be matched to activate
        the bind, defaults to None
        :return: a dict containing the previous joystick data
        """

    def rebind(
        self,
        key_bind_name: str,
        *args,
        **kwds,
    ) -> dict | tuple[int | None, int | None] | None:
        new_bind: Any
        if len(args):
            new_bind = args[0]
        elif new_bind := kwds.get("new_key"):
            pass
        if kwds.get("new_joystick_data") or isinstance(new_bind, dict):
            return self._rebind_joystick(key_bind_name, new_bind)
        else:
            mod_keys: int | None
            if len(args) > 1:
                mod_keys = args[1]
            else:
                mod_keys = kwds.get("new_mod", None)
            return self._rebind_key(key_bind_name, new_bind, mod_keys)

    def _rebind_key(
        self,
        key_bind_name: str,
        new_key: Optional[int] = None,
        new_mod: Optional[int] = None,
    ) -> tuple[int | None, int | None] | None:
        old_bind: tuple | None
        try:
            old_bind = self.key_map.get_bound_key(key_bind_name)
        except ValueError:
            logger.warning(
                f"Attempted to rebind '{key_bind_name}' when bind does not"
                " exist. \n Program might be attempting to rebind before"
                " generating binds, or bind name may be incorrect."
            )
            return None
        self.key_map.rebind(
            KeyBind(bind_name=key_bind_name, mod=new_mod), new_key=new_key
        )

        return old_bind

    def _rebind_joystick(
        self,
        key_bind_name: str,
        new_joystick_data: Optional[dict] = None,
    ) -> dict | None:
        old_bind: dict | None
        try:
            old_bind = self.joy_map.get_bound_joystick_event(key_bind_name)
        except ValueError:
            logger.warning(
                f"Attempted to rebind '{key_bind_name}' when bind does not"
                " exist. \n Program might be attempting to rebind before"
                " generating binds, or bind name may be incorrect."
            )
            return None
        self.joy_map.rebind(key_bind_name, new_joystick_data)

        return old_bind

    def unbind(self, func: Callable, bind_name: Optional[str] = None) -> None:
        """
        Removes a callable from the given bind. Removes the callable from all binds,
        if none is given

        :param func: A Callable previously registered with this Key Listener
        :param bind_name: The bind to be removed from, or all instances, if
        None. Defaults to None.
        """
        if bind_name:
            event_dict = self._key_hooks.get(bind_name)
            if not event_dict:
                logger.warning(
                    f"Bind name '{bind_name}' does not exist in KeyListener "
                    f"'{self.handle}'"
                )
                return
            found = False
            for bind_dict in event_dict.values():
                for bind in bind_dict.values():
                    if func not in bind:
                        continue
                    found = True
                    bind.remove(func)
            if not found:
                logger.warning(
                    f"Cannot remove function {func.__name__} from '"
                    f"{bind_name}' of KeyListener: {self.handle}.\n"
                    f"Function is not bound to that name."
                )
            return
        for event_dict in self._key_hooks.values():
            for bind_dict in event_dict.values():
                if not bind_dict:
                    continue
                for bind in bind_dict.values():
                    if bind and func in bind:
                        bind.remove(func)

    def _capture_method(
        self, cls: Type[object], method: Callable, tag_data: tuple
    ) -> None:
        """
        Adds the method, class, and event into the appropriate dictionaries to ensure
        they can be properly notified.

        :param cls: Class of the object being processed
        :param method: Callable being captured
        :param tag_data: A tuple containing pertinent registration data
        """
        is_concurrent = not hasattr(method, "_runs_sequential")
        key_bind_name: str = tag_data[0]
        default_key: int = tag_data[1]
        default_mod: int = tag_data[2]
        event_type: int = tag_data[3]
        default_joystick_data: dict = tag_data[4]

        if default_joystick_data is not None:
            self.joy_map.generate_bind(key_bind_name, default_joystick_data)
        else:
            self.key_map.generate_bind(key_bind_name, default_key, default_mod)

        # -----Add to Class Listeners-----
        event_dict = self._class_listeners.setdefault(key_bind_name, {})
        concurrency_dict = event_dict.setdefault(is_concurrent, {})
        listeners = concurrency_dict.setdefault(event_type, [])
        listeners.append((method, cls))

        # -----Add to Class Listener Events-----

        self._class_listener_binds.setdefault(method, []).append(key_bind_name)

        # -----Add to Assigned Classes-----
        self._assigned_classes.setdefault(cls, []).append(method)

    def bind_method(
        self,
        key_bind_name: str,
        default_key: Optional[int] = None,
        default_mod: Optional[int] = None,
        default_joystick_data: Optional[dict] = None,
        event_type: int = pygame.KEYDOWN,
    ) -> Callable:
        """
        Wrapper that marks the method for registration when the class is registered.

        The method's class should be registered with all event managers that have
        registered a method in that class. Failure to do so will leave a dangling
        attribute on those methods.

        :param event_type: Pygame event type that will call the assigned method.
        """

        def decorator(method: Callable) -> Callable:
            return self._tag_method(
                method,
                (
                    key_bind_name,
                    default_key,
                    default_mod,
                    event_type,
                    default_joystick_data,
                ),
            )

        return decorator

    def deregister_class(self, cls: Type[object]):
        """
        Clears all instances and methods belonging to the supplied class.

        :param cls: The cls being deregistered.
        :raises KeyError: If cls is not contained in the class listeners, this
        error will be raised.
        """
        self._class_listener_instances.pop(cls, None)
        for method in self._assigned_classes.get(cls, []):
            self.unbind_method(method)
        self._assigned_classes.pop(cls)

    def unbind_method(self, method: Callable):
        """
        Clears the method from its bindings.

        :param method: Method being unbound
        """
        for bind_name in self._class_listener_binds.get(method, []):
            concurrency_dict = self._class_listeners.get(bind_name, {})
            for concurrency, event_dict in concurrency_dict.items():
                for event_type, listener_sets in event_dict.items():
                    # Retain only the listeners that are not the method
                    listener_sets = list(
                        filter(
                            lambda listener_set: method is not listener_set[0],
                            listener_sets,
                        )
                    )
                    event_dict.update({event_type: listener_sets})
                concurrency_dict.update({concurrency: event_dict})
            self._class_listeners.update({bind_name: concurrency_dict})
        self._class_listener_binds.pop(method)

    def clear_bind(self, bind_name: str, eliminate_bind: bool = False) -> None:
        """
        Clears all callables from the specified bind name

        :param bind_name: he name of the bind being cleared
        :param eliminate_bind: Boolean for determining if the bind should be removed
        from the JoyMap and KeyMap, defaults to False
        """
        if eliminate_bind:
            bind = self._key_hooks.pop(bind_name, None)
            if bind is None:
                logger.warning(
                    f" Cannot remove bind '{bind_name}';" " bind does not exist."
                )
                return
            self.key_map.remove_bind(bind_name)
            self.joy_map.remove_bind(bind_name)
            return
        call_list = self._key_hooks.get(bind_name, None)
        class_call_list = self._class_listeners.get(bind_name, None)
        if call_list is None and class_call_list is None:
            logger.warning(f" Bind '{bind_name}' not in key registry.")
            return
        if call_list:
            logger.info(f"Clearing all functions from bind {bind_name}")
            call_list.clear()
        if class_call_list:
            logger.info(f"Clearing all methods from bind {bind_name}")
            class_call_list.clear()

    def _validate_input(self, key_bind: KeyBind, event: pygame.Event) -> bool:
        """
        Validates the input data against a key bind to ensure a match

        :param key_bind: Target key bind containing desired input data
        :param input_data: Data coming from the recent input event
        :return: True if the input data matches the bind, otherwise false
        """
        is_valid = False
        # This structure is here to potentially leave room for
        key_changed: int | None = getattr(event, "key", None)
        mod_keys: int | None = getattr(event, "mod", None)
        if key_changed is not None:
            mod = key_bind.mod
            if mod is None:
                is_valid = True
            elif mod_keys is not None and (mod & mod_keys or mod is mod_keys):
                # mod is mod_keys catches pygame.KMOD_NONE
                is_valid = True

        return is_valid

    def _get_callables(self, event: pygame.Event) -> _CallableSets:
        """
        Calls all registered functions and methods that make use of the given event

        :param event: pygame event to be passed to the callables
        """
        key_changed: int | None = getattr(event, "key", None)
        # mod_keys: int | None = getattr(event, "mod", None)

        # input_data = (key_changed, mod_keys)
        conc_funcs_lists = []
        seq_funcs_lists = []
        conc_methods_lists = []
        seq_methods_lists = []
        for key_bind in self.key_map.key_binds.get(key_changed, []):
            # Try to match the mod keys. If they don't, move on to the next.
            if not self._validate_input(key_bind, event):
                continue
            hooks = self._key_hooks.get(key_bind.bind_name, {})
            conc_funcs_lists.append(hooks.get(True, {}).get(event.type, []))
            seq_funcs_lists.append(hooks.get(False, {}).get(event.type, []))

            method_hooks = self._class_listeners.get(key_bind.bind_name, {})
            conc_methods_lists.append(method_hooks.get(True, {}).get(event.type, []))
            seq_methods_lists.append(method_hooks.get(False, {}).get(event.type, []))
        for bind in self.joy_map.get(event, []):
            hooks = self._key_hooks.get(bind, {})
            conc_funcs_lists.append(hooks.get(True, {}).get(event.type, []))
            seq_funcs_lists.append(hooks.get(False, {}).get(event.type, []))

            method_hooks = self._class_listeners.get(bind, {})
            conc_methods_lists.append(method_hooks.get(True, {}).get(event.type, []))
            seq_methods_lists.append(method_hooks.get(False, {}).get(event.type, []))
        return _CallableSets(
            concurrent_functions=list(itertools.chain(*conc_funcs_lists)),
            sequential_functions=list(itertools.chain(*seq_funcs_lists)),
            concurrent_methods=list(itertools.chain(*conc_methods_lists)),
            sequential_methods=list(itertools.chain(*seq_methods_lists)),
        )

    @classmethod
    def load_from_file(
        cls, file_path: PathLike | str, parser: Optional[Type[FileParser]] = None
    ) -> None:
        """
        Pulls the file from the file path, and uses the supplied parser to convert the
        file into a KeyMap, which is merged with the current KeyMap.

        Binds in the current KeyMap that don't exist in the loaded KeyMap do not change,
        all others are updated to reflect the loaded binds

        If a FileParser subclass is not included, it will be inferred from the file
        type. If the file type is not recognized/supported, it will generate an error.

        :param file_path: Path to the file to be loaded
        :param parser: Parser to be used to decode the file, defaults to None
        :raises ValueError: Raised if not parser is given, and the file type is not
        supported.
        """
        if type(file_path) is not Path:
            file_path = Path(file_path)
        parser = parser or _get_parser_from_path(file_path)
        with open(file_path, "r") as file:
            key_binds, joy_binds = parser.load(file)
            cls.key_map.merge(key_binds)
            cls.joy_map.merge(joy_binds)

    @classmethod
    def save_to_file(
        cls, file_path: PathLike | str, parser: Optional[Type[FileParser]] = None
    ) -> None:
        """
        Saves the current KeyMap and JoyMap to a file in the requested location.
        Requires a path object that includes the file type.

        If a FileParser subclass is not included, it will be inferred from the file
        type. If the file type is not recognized/supported, it will generate an error.

        :param file_path: Path to the file being saved to
        :param parser: Parser used to encode the file, defaults to None
        :raises ValueError: Raised if not parser is given, and the file type is not
        supported.
        """
        if type(file_path) is not Path:
            file_path = Path(file_path)
        parser = parser or _get_parser_from_path(file_path)
        with open(file_path, "w") as file:
            parser.save(cls.key_map, cls.joy_map, file)


def notifyKeyListeners(event: pygame.Event) -> None:
    """
    Automatically passes the event to all existing KeyListeners

    :param event: Pygame event instance, of type KEYDOWN or KEYUP
    """
    for listener in KeyListener._listeners.values():
        listener.notify(event)


def getKeyListener(handle: str) -> KeyListener:
    """
    Supplies a Key Listener with the given handle. If one exists with that handle,
    the existing Key Listener is given. Otherwise, a new one is created.

    :param handle: String describing the KeyListener.
    """
    return KeyListener._listeners.setdefault(handle, KeyListener(handle))



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Optional, overload

import pygame


# These are the types of event data we don't care about
# 'value' is a read out value, and typically analog, so we don't want that captured
# 'instance_id" is the specific joystick instace, we want to call regardless of that
# 'joy' is deprecated and can safely be ignored
_invalid_params = ["value", "instance_id", "joy"]


@dataclass
class JoyMap:
    """
    Gets a bind name from joystick event parameters
    """

    _joy_binds: dict[tuple[tuple] | None, list[str]] = field(default_factory=dict)

    @overload
    def _convert_event(self, event: dict) -> tuple[tuple]: ...

    @overload
    def _convert_event(self, event: pygame.Event) -> tuple[tuple]: ...

    def _convert_event(self, event: Any) -> tuple[tuple]:
        """
        Converts the dict of an event into a tuple of tuple suitable to use as a dict
        key.

        Ignores unneeded parameters

        :param event: pygame event, or a dict containing the data of an event.
        :return: The event's dict, converted into a tuple of tuples
        """
        event_dict: dict
        arg_type = type(event)
        if arg_type is pygame.Event:
            event_dict = event.__dict__
        elif arg_type is dict:
            event_dict = event
        else:
            raise ValueError("Invalid argument type")
        pairs: list[tuple] = []
        for key, value in event_dict.items():
            if key in _invalid_params:
                continue
            pairs.append((key, value))
        # This is a valid conversion, but Mypy doesn't like it
        return tuple(pairs)  # type: ignore

    def _convert_pairs(self, event_key: tuple[tuple]) -> dict:
        return dict((key, value) for key, value in event_key)

    def get(self, event: pygame.Event, default: Optional[list] = None) -> list[str]:
        """
        Returns the list of bind names that match the given event.

        :param event: Event generated by joystick input
        :param default: Return value if no binds match the event, defaults to None
        :return: A list of bind names, or the given default value, or None
        :raises ValueError: Raised if the bind name is not found.
        """
        # Convert the dict of an event into a tuple of tuple to use as a dict key
        key = self._convert_event(event)
        if default is None:
            default = []
        return self._joy_binds.get(key, default)

    def get_bound_joystick_event(self, bind_name: str) -> dict | None:
        """
        Returns joystick event data that a bind is tied to.
        If a bind exists but is unbound, returns None

        :param bind_name: Name of target bind
        :raises ValueError: If bind_name is not found
        :return: dict containing joystick event data, or None if unbound
        """
        for key, bind_list in self._joy_binds.items():
            if bind_name in bind_list:
                if key is None:
                    return key
                return self._convert_pairs(key)
        raise ValueError(f"Bind name: {bind_name} not found.")

    def generate_bind(
        self, bind_name: str, default_joystick_data: Optional[dict] = None
    ) -> None:
        try:
            self.get_bound_joystick_event(bind_name)
        except ValueError:
            self._joy_binds.setdefault(
                # adds the bind to joy_binds, under the given parameters
                (
                    self._convert_event(default_joystick_data)
                    if default_joystick_data is not None
                    else None
                ),
                [],
            ).append(bind_name)

    def _rebind(
        self, bind_name: str, new_joystick_data: Optional[tuple[tuple]] = None
    ) -> None:
        self.remove_bind(bind_name)
        self._joy_binds.setdefault(new_joystick_data, []).append(bind_name)

    def rebind(self, bind_name: str, new_joystick_data: Optional[dict] = None):
        """
        Converts a bind from its current expected data to the new data for matching.

        :param bind_name: he name of the bind whose calling event is changing.
        :param new_joystick_data: The new joystick event data, defaults to None.
        If none, the bind name is considered unbound, and will not be called by notify()
        """
        self._rebind(
            bind_name,
            (
                self._convert_event(new_joystick_data)
                if new_joystick_data is not None
                else None
            ),
        )

    def remove_bind(self, bind_name: str) -> None:
        """
        Clears the given bind from all joystick inputs.

        :param bind_name: Name of the target bind
        """
        dead_keys = []
        for key, bind_list in self._joy_binds.items():
            if bind_name in bind_list:
                bind_list.remove(bind_name)
            if len(bind_list) == 0:
                dead_keys.append(key)
        for key in dead_keys:
            self._joy_binds.pop(key)

    def merge(self, other: JoyMap) -> None:
        """
        Combines two JoyBinds, adding any new binds from the latter and updating
        overlapping binds to match the latter, while preserving any unique binds

        :param other: A JoyMap with preferred binds.
        """
        for joy_data, bind_list in other._joy_binds.items():
            for bind in bind_list:
                self._rebind(bind, joy_data)

    def pack_binds(self) -> dict:
        """
        Converts the dictionary of binds and their needed inputs into a saveable dict,
        with the bind name forward for manual remapping in a file
        """
        packed_dict: dict[str, tuple[tuple] | None] = {}
        for joy_data, bind_list in self._joy_binds.items():
            for bind in bind_list:
                packed_dict.update({bind: joy_data})

        return packed_dict



from __future__ import annotations
from dataclasses import dataclass, field
import logging
from typing import NamedTuple, Optional

import pygame

logger: logging.Logger = logging.getLogger(__name__)

KeyBind = NamedTuple("KeyBind", [("bind_name", str), ("mod", int | None)])


@dataclass
class KeyMap:
    """
    A dictionary of keybind events with associated keys.
    """

    key_binds: dict[int | None, list[KeyBind]] = field(default_factory=dict)

    def rebind(self, new_key_bind: KeyBind, new_key: Optional[int] = None) -> None:
        """
        Takes the given keybind, unbinds it from its current key, and rebinds
        it to the given key. If no key is given, it is put under None, meaning
        the hook is regarded as unbound, and cannot be called.

        :param new_key_bind: Keybind containing the bind name and mod keys.
        :param new_key: Pygame key id, defaults to None
        """
        dead_keys: set[int | None] = set()
        for key, key_bind_list in self.key_binds.items():
            binds_to_remove: list[KeyBind] = []
            key_bind: KeyBind
            for key_bind in key_bind_list:
                if key_bind.bind_name == new_key_bind.bind_name:
                    # Can't directly remove in the loop, could skip one.
                    binds_to_remove.append(key_bind)
            removed_bind: KeyBind
            for removed_bind in binds_to_remove:
                key_bind_list.remove(removed_bind)
            if not key_bind_list:
                dead_keys.add(key)
        # Clean up any empty lists
        for key in dead_keys:
            self.key_binds.pop(key)

        self.key_binds.setdefault(new_key, []).append(new_key_bind)

    def get_bound_key(self, bind_name: str) -> tuple[int | None, int | None]:
        """
        Finds the current key that binds to the given bind name, and
        returns it in a tuple with the bound mod keys.

        :param bind_name: Name of the bind being used.
        :return: Tuple containing two ints, first representing the number for
        a pygame key, the second a bitmask int representing pygame mod keys.
        :raises ValueError: Raised if the bind name is not found.
        """
        key: int | None
        key_bind_list: list[KeyBind]
        key_bind: KeyBind
        for key, key_bind_list in self.key_binds.items():
            for key_bind in key_bind_list:
                if key_bind.bind_name == bind_name:
                    return key, key_bind.mod
        raise ValueError(f"Bind name: {bind_name} not found.")

    def generate_bind(
        self,
        key_bind_name: str,
        default_key: Optional[int] = None,
        default_mod: Optional[int] = None,
    ) -> None:
        """
        Looks for a bind matching the given name.
        Creates the bind if it does not exist.
        Does not overwrite the key of an existing bind.

        :param key_bind_name: Name assigned to the bind.
        :param default_key: Pygame key code assigned to a new bind,
        defaults to None
        :param default_mod: bitmask of pygame modkeys for a new bind,
        defaults to pygame.KMOD_NONE
        """
        try:
            self.get_bound_key(key_bind_name)
        except ValueError:
            self.key_binds.setdefault(default_key, []).append(
                KeyBind(bind_name=key_bind_name, mod=default_mod)
            )

    def remove_bind(self, bind_name: str, key: Optional[int] = None) -> None:
        """
        Eliminates the specified bind from the specified key, or all instances
        if no key is specified.

        :param bind_name: Name of the bind being removed
        :param key: Integer of pygame key code to search, defaults to None
        """
        key_bind_list: list[KeyBind] | None
        if key is not None:
            key_bind_list = self.key_binds.get(key, None)
            if not key_bind_list:
                logger.warning(
                    f" Cannot remove '{bind_name}';"
                    f" {pygame.key.name(key)} does not have any binds."
                )
                return
            to_remove: list[KeyBind] = []
            for key_bind in key_bind_list:
                if key_bind.bind_name == bind_name:
                    to_remove.append(key_bind)
            for item in to_remove:
                key_bind_list.remove(item)
            if not to_remove:
                logger.warning(
                    f" Cannot remove '{bind_name}';"
                    f" bind does not exist in {pygame.key.name(key)}"
                )
            return
        for key_bind_list in self.key_binds.values():
            to_remove = []
            for key_bind in key_bind_list:
                if key_bind.bind_name == bind_name:
                    to_remove.append(key_bind)
            for item in to_remove:
                key_bind_list.remove(item)
        return None

    def merge(self, other: KeyMap) -> None:
        """
        Uses the supplied other map to rebind any existing keys to match the other map.
        Adds any binds from other map into the current key map, and preserves any other
        binds it already holds.

        :param other: Key map to be merged.
        """
        for key, bind_list in other.key_binds.items():
            for bind in bind_list:
                self.rebind(bind, key)

    def pack_binds(self) -> dict:
        """
        Packages binds into a serializable dictionary that is more easily saved.
        """
        packed_dict: dict[str, tuple[str | None, int | None]] = {}
        for key_code, bind_list in self.key_binds.items():
            for bind in bind_list:
                key_name = None
                if key_code:
                    key_name = pygame.key.name(key_code)
                packed_dict.update({bind.bind_name: (key_name, bind.mod)})

        return packed_dict



from __future__ import annotations

import logging
from typing import Callable, Optional, Type

from .base_manager import BaseManager, _CallableSets

import pygame

logger: logging.Logger = logging.getLogger(__name__)


class EventManager(BaseManager):
    handlers: dict[str, EventManager] = {}

    def __init__(self, handle: str) -> None:
        super().__init__(handle)

        # --------Basic function assignment--------
        # Pygame event as key, list of functions as values
        self._listeners: dict[int, dict[bool, list[Callable]]] = {}

        # --------Class method assignment--------
        # Pygame event key, method and affected object as values
        self._class_listeners: dict[
            int, dict[bool, list[tuple[Callable, Type[object]]]]
        ] = {}
        # Inversion of _class_listeners. Method as key, event id as values
        self._class_listener_events: dict[Callable, list[int]] = {}

    def register(self, event_type: int) -> Callable:
        """
        Takes a callable item such as a function, and places it in the appropriate set
        of functions with the target event. Whenever the event manager is notified of
        the compatible event type, it will call the function.

        :param event_type: Pygame event code
        :return: Returns the registered callable, unchanged.
        """

        def decorator(listener: Callable) -> Callable:
            is_concurrent = not hasattr(listener, "_runs_sequential")
            event_dict = self._listeners.setdefault(event_type, {})
            concurrency_list = event_dict.setdefault(is_concurrent, [])
            concurrency_list.append(listener)
            return listener

        return decorator

    def deregister(self, func: Callable, event_type: Optional[int] = None) -> None:
        """
        Remove the given function from the specified event type. If no event
        type is specified, the function is cleared from all events.

        :param func: Function to be removed from the register.
        :param event_type: Pygame event type to which the function is to be
        removed, defaults to None
        """
        call_list: list[Callable] | None
        if event_type is not None:
            event_dict = self._listeners.get(event_type)
            if not event_dict:
                logger.warning(
                    "No functions are registered to "
                    f"{pygame.event.event_name(event_type)}"
                )
                return
            found = False
            for call_list in event_dict.values():
                if func not in call_list:
                    continue
                found = True
                call_list.remove(func)
            if not found:
                logger.warning(
                    f"Function '{func.__name__}' is not bound to "
                    f"{pygame.event.event_name(event_type)}"
                )
            return
        for event_dict in self._listeners.values():
            if not event_dict:
                continue
            for call_list in event_dict.values():
                if func in call_list:
                    call_list.remove(func)

    def _capture_method(self, cls, method, tag_data):
        """
        Adds the method, class, and event into the appropriate dictionaries to ensure
        they can be properly notified.

        :param cls: Class of the object being processed
        :param method: Callable being captured
        :param tag_data: A tuple containing pertinent registration data
        """
        is_concurrent = not hasattr(method, "_runs_sequential")
        event_type = tag_data[0]  # Only piece of data

        # -----Add to Class Listeners-----
        event_dict = self._class_listeners.setdefault(event_type, {})
        concurrency_list = event_dict.setdefault(is_concurrent, [])
        concurrency_list.append((method, cls))

        # -----Add to Class Listener Events-----
        self._class_listener_events.setdefault(method, []).append(event_type)

        # -----Add to Assigned Classes-----
        self._assigned_classes.setdefault(cls, []).append(method)

    def register_method(self, event_type: int) -> Callable:
        """
        Wrapper that marks the method for registration when the class is registered.

        The method's class should be registered with all event managers that have
        registered a method in that class. Failure to do so will leave a dangling
        attribute on those methods.

        :param event_type: Pygame event type that will call the assigned method.
        """

        def decorator(method: Callable) -> Callable:
            return self._tag_method(method, (event_type,))

        return decorator

    def deregister_class(self, cls: Type[object]):
        """
        Clears all instances and listeners that belong to the supplied class.

        :param cls: The cls being deregistered.
        :raises KeyError: If cls is not contained in the class listeners, this
        error will be raised.
        """
        # Purge instances
        self._class_listener_instances.pop(cls, None)
        # Remove methods from events
        for method in self._assigned_classes.get(cls, []):
            self.deregister_method(method)
        self._assigned_classes.pop(cls)

    def deregister_method(self, method: Callable):
        """
        Clears the method from the registry so it is no longer called when the assigned
        event is fired.

        :param method: Method whose registration is being revoked.
        """
        for event_type in self._class_listener_events.get(method, []):
            event_dict = self._class_listeners.get(event_type, {})
            for concurrency, listener_sets in event_dict.items():
                # Retain only the listeners that are not the method
                listener_sets = list(
                    filter(
                        lambda listener_set: method is not listener_set[0],
                        listener_sets,
                    )
                )
                event_dict.update({concurrency: listener_sets})
            self._class_listeners.update({event_type: event_dict})
        self._class_listener_events.pop(method)

    def purge_event(self, event_type: int) -> None:
        """
        Attempts to clear all functions from the specified event.

        :param event_type: Pygame event type
        """
        self._listeners.pop(event_type, None)
        self._class_listeners.pop(event_type, None)
        # This really simplified things, no?

    def _get_callables(self, event) -> _CallableSets:
        functions = self._listeners.get(event.type, {})
        methods = self._class_listeners.get(event.type, {})
        return _CallableSets(
            concurrent_functions=functions.get(True, []),
            sequential_functions=functions.get(False, []),
            concurrent_methods=methods.get(True, []),
            sequential_methods=methods.get(False, []),
        )


def notifyEventManagers(event: pygame.Event) -> None:
    """
    Passes on the event to all existing EventManagers.

    :param event: Pygame-generated event that is being handled.
    """
    for event_handler in EventManager.handlers.values():
        event_handler.notify(event)


def getEventManager(handle: str) -> EventManager:
    """
    Finds the handler that matches the given handle.
    If one does not exist, it is created.

    :param handle: A string for identifying an event manager instance.
    """
    return EventManager.handlers.setdefault(handle, EventManager(handle))



from io import StringIO
import json
import pathlib
import sys
import threading
from typing import Callable, cast, Type
import unittest

import pygame

sys.path.append(str(pathlib.Path.cwd()))

from src.simple_events.key_manager import (  # noqa: E402
    getKeyListener,
    KeyListener,
)

from src.simple_events.key_map import (  # noqa: E402
    KeyBind,
    KeyMap,
)

from src.simple_events.joy_map import (  # noqa: E402
    JoyMap,
)

from src.simple_events.file_parser import (  # noqa: E402
    _get_parser_from_path,
    JSONParser,
)


class TestKeyMap(unittest.TestCase):

    def setUp(self) -> None:
        self.keymap = KeyMap()

    def tearDown(self) -> None:
        self.keymap.key_binds.clear()

    def test_generate_bind(self):
        bind_name = "test_bind"
        key = pygame.K_9
        mod = pygame.KMOD_ALT

        self.keymap.generate_bind(bind_name, key, mod)

        test_bind = KeyBind(bind_name, mod)
        self.assertIn(test_bind, self.keymap.key_binds.get(key, []))

    def test_rebind(self) -> None:
        bind_name = "test_bind"
        start_key = pygame.K_9
        start_mod = pygame.KMOD_ALT
        test_bind = KeyBind(bind_name=bind_name, mod=start_mod)

        self.keymap.generate_bind(bind_name, start_key, start_mod)

        new_key = pygame.K_0

        self.keymap.rebind(test_bind, new_key)

        self.assertNotIn(test_bind, self.keymap.key_binds.get(start_key, []))
        self.assertIn(test_bind, self.keymap.key_binds.get(new_key, []))

    def test_get_bound_key(self) -> None:
        bind_name = "test_bind"
        start_key = pygame.K_9
        start_mod = pygame.KMOD_ALT

        self.keymap.generate_bind(bind_name, start_key, start_mod)

        key, mod = cast(tuple[int | None, int], self.keymap.get_bound_key(bind_name))
        self.assertRaises(ValueError, self.keymap.get_bound_key, "unbound_name")
        self.assertEqual(key, start_key)
        self.assertEqual(start_mod, mod)

    def test_remove_bind(self) -> None:
        bind_name = "test_bind"
        start_key = pygame.K_9
        start_mod = pygame.KMOD_ALT
        test_bind = KeyBind(bind_name=bind_name, mod=start_mod)

        self.keymap.generate_bind(bind_name, start_key, start_mod)

        new_key = pygame.K_0

        self.keymap.key_binds.setdefault(new_key, []).append(test_bind)

        self.keymap.remove_bind(bind_name, start_key)

        self.assertNotIn(
            test_bind, cast(list[KeyBind], self.keymap.key_binds.get(start_key))
        )
        self.assertIn(
            test_bind, cast(list[KeyBind], self.keymap.key_binds.get(new_key))
        )

        self.keymap.remove_bind(bind_name)

        self.assertNotIn(
            test_bind, cast(list[KeyBind], self.keymap.key_binds.get(start_key))
        )
        self.assertNotIn(
            test_bind, cast(list[KeyBind], self.keymap.key_binds.get(new_key))
        )

    def test_merge(self) -> None:
        other_map = KeyMap()

        for i in range(3):
            self.keymap.generate_bind(f"bind{i}", pygame.K_0)

        for i in range(2, 5):
            other_map.generate_bind(f"bind{i}", pygame.K_1)

        self.keymap.merge(other_map)

        expected_key = pygame.K_0
        try:
            for i in range(5):
                key = pygame.K_0 if i < 2 else pygame.K_1
                expected_key = key
                self.assertEqual(key, self.keymap.get_bound_key(f"bind{i}")[0])
        except ValueError:
            self.fail(
                f"Expected key {pygame.key.name(expected_key)}, "
                f"but couldn't find 'bind{i}'"
            )

    def test_pack_binds(self) -> None:

        for i in range(3):
            self.keymap.generate_bind(f"bind{i}", pygame.K_0)
        self.keymap.generate_bind("bind3", None)

        packed_dict = self.keymap.pack_binds()

        comp_dict = {
            "bind0": ("0", None),
            "bind1": ("0", None),
            "bind2": ("0", None),
            "bind3": (None, None),
        }

        self.assertDictEqual(packed_dict, comp_dict)


class TestJoyMap(unittest.TestCase):

    def setUp(self):
        self.joymap = JoyMap()

    def tearDown(self):
        self.joymap._joy_binds.clear()

    def test_convert_event(self):
        joystick_data = {"button": 0}
        converted_data = self.joymap._convert_event(joystick_data)

        match_data = (("button", 0),)

        self.assertEqual(converted_data, match_data)

        event = pygame.Event(pygame.JOYBUTTONDOWN, button=0, instance_id=0, joy=0)

        converted_data = self.joymap._convert_event(event)

        self.assertEqual(converted_data, match_data)

        event = pygame.Event(
            pygame.JOYAXISMOTION, axis=0, instance_id=0, joy=0, value=0.5
        )

        match_data = (("axis", 0),)

        converted_data = self.joymap._convert_event(event)

        self.assertEqual(converted_data, match_data)

        event = pygame.Event(
            pygame.JOYHATMOTION, hat=0, instance_id=0, joy=0, value=(0, 0)
        )

        match_data = (("hat", 0),)

        converted_data = self.joymap._convert_event(event)

        self.assertEqual(converted_data, match_data)

    def test_convert_pairs(self):
        key_data = (("button", 0),)
        converted_data = self.joymap._convert_pairs(key_data)

        match_data = {"button": 0}

        self.assertEqual(converted_data, match_data)

    def test_generate_bind(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        match_data = (("button", 0),)

        self.assertIn(bind_name, self.joymap._joy_binds.get(match_data, []))

    def test_get(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        event = pygame.Event(pygame.JOYBUTTONDOWN, button=0, instance_id=0, joy=0)

        bind_list = self.joymap.get(event)

        self.assertIn(bind_name, bind_list)

    def test_get_bound_joystick_event(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        returned_data = self.joymap.get_bound_joystick_event(bind_name)

        self.assertEqual(returned_data, joystick_data)

    def test_remove_bind(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        match_data = (("button", 0),)

        self.assertIn(bind_name, self.joymap._joy_binds.get(match_data, []))

        self.joymap.remove_bind(bind_name)

        self.assertNotIn(bind_name, self.joymap._joy_binds.get(match_data, []))

    def test_rebind(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        match_data = (("button", 0),)

        self.assertIn(bind_name, self.joymap._joy_binds.get(match_data, []))

        new_joystick_data = {"button": 1}

        new_match_data = (("button", 1),)

        self.joymap.rebind(bind_name, new_joystick_data)

        self.assertNotIn(bind_name, self.joymap._joy_binds.get(match_data, []))

        self.assertIn(bind_name, self.joymap._joy_binds.get(new_match_data, []))

    def test_merge(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        match_data1 = (("button", 0),)

        bind_name2 = "test_bind2"
        joystick_data2 = {"button": 0}
        self.joymap.generate_bind(bind_name2, joystick_data2)

        self.assertIn(bind_name, self.joymap._joy_binds.get(match_data1, []))
        self.assertIn(bind_name2, self.joymap._joy_binds.get(match_data1, []))

        other_map = JoyMap()

        bind_name3 = "test_bind2"
        joystick_data3 = {"button": 1}
        other_map.generate_bind(bind_name3, joystick_data3)

        self.joymap.merge(other_map)

        new_match_data = (("button", 1),)

        self.assertIn(bind_name, self.joymap._joy_binds.get(match_data1, []))
        self.assertNotIn(bind_name2, self.joymap._joy_binds.get(match_data1, []))
        self.assertIn(bind_name2, self.joymap._joy_binds.get(new_match_data, []))

    def test_pack_binds(self):
        bind_name = "test_bind"
        joystick_data = {"button": 0}
        self.joymap.generate_bind(bind_name, joystick_data)

        bind_name2 = "test_bind2"
        joystick_data2 = {"button": 0}
        self.joymap.generate_bind(bind_name2, joystick_data2)

        test_dict = {"test_bind": (("button", 0),), "test_bind2": (("button", 0),)}

        packed_dict = self.joymap.pack_binds()

        self.assertEqual(test_dict, packed_dict)


class TestFileParser(unittest.TestCase):

    def test_json(self):
        path = pathlib.PurePath("sample.json")

        self.assertIs(_get_parser_from_path(path), JSONParser)

        new_path = pathlib.PurePath("unsupported.py")

        with self.assertRaises(ValueError):
            _get_parser_from_path(new_path)


class TestJSONParser(unittest.TestCase):

    def test_unpack_keys(self) -> None:

        keymap = KeyMap()

        for i in range(3):
            keymap.generate_bind(f"bind{i}", pygame.K_0)
        keymap.generate_bind("bind3", None)

        packed = keymap.pack_binds()

        unpacked = JSONParser._unpack_keys(packed)

        comp_dict = {
            pygame.K_0: [
                KeyBind("bind0", None),
                KeyBind("bind1", None),
                KeyBind("bind2", None),
            ],
            None: [("bind3", None)],
        }

        self.assertDictEqual(unpacked, comp_dict)

    def test_save(self) -> None:

        keymap = KeyMap()
        joymap = JoyMap()

        for i in range(3):
            keymap.generate_bind(f"bind{i}", pygame.K_0)
        keymap.generate_bind("bind3", None)

        outfile = StringIO()

        JSONParser.save(keymap, joymap, outfile)

        outfile.seek(0)

        maps = {"keys": keymap.pack_binds(), "controller": joymap.pack_binds()}

        json_string = json.dumps(maps, indent=2)

        out_string = outfile.read()

        self.assertEqual(out_string, json_string)

    def test_load(self) -> None:

        keymap = KeyMap()

        for i in range(3):
            keymap.generate_bind(f"bind{i}", pygame.K_0)
        keymap.generate_bind("bind3", None)

        json_string = (
            r'{"keys": '
            r'{"bind0": ["0", null], "bind1": ["0", null], '
            r'"bind2": ["0", null], "bind3": [null, null]}, '
            r'"controller": {}'
            r"}"
        )

        infile = StringIO()
        infile.write(json_string)
        infile.seek(0)

        new_map, _ = JSONParser.load(infile)

        for key in new_map.key_binds:
            self.assertEqual(new_map.key_binds.get(key), keymap.key_binds.get(key))


class TestKeyListener(unittest.TestCase):

    def assertHasAttr(self, obj, intendedAttr: str):
        testBool = hasattr(obj, intendedAttr)

        self.assertTrue(testBool, msg=f"{obj=} lacks an attribute, {intendedAttr=}")

    def assertNotHasAttr(self, obj, intendedAttr: str):
        testBool = hasattr(obj, intendedAttr)

        self.assertFalse(
            testBool, msg=f"{obj=} has unexpected attribute, {intendedAttr=}"
        )

    def setUp(self) -> None:
        self.key_listener = getKeyListener("TestCase")

    def tearDown(self) -> None:
        self.key_listener.key_map.key_binds.clear()
        self.key_listener._key_hooks.clear()
        self.key_listener._assigned_classes.clear()
        self.key_listener._class_listener_binds.clear()
        self.key_listener._class_listeners.clear()
        self.key_listener._class_listener_instances.clear()

    def test_sequential_tag(self) -> None:

        @self.key_listener.bind("test_bind1")
        @self.key_listener.sequential
        def test_func() -> None:
            pass

        self.assertHasAttr(test_func, "_runs_sequential")

    def test_concurrent_tag(self) -> None:

        @self.key_listener.bind("test_bind1")
        @self.key_listener.concurrent
        @self.key_listener.sequential
        def test_func() -> None:
            pass

        self.assertNotHasAttr(test_func, "_runs_sequential")

    def test_bind(self) -> None:

        def test_func(_) -> None:
            pass

        self.key_listener.bind("test_bind0", pygame.K_0, pygame.KMOD_ALT)(test_func)

        self.key_listener.bind("test_bind1", pygame.K_1)(test_func)

        self.key_listener.bind("test_bind2")(test_func)

        self.assertIn("test_bind0", self.key_listener._key_hooks.keys())
        self.assertIn("test_bind1", self.key_listener._key_hooks.keys())
        self.assertIn("test_bind2", self.key_listener._key_hooks.keys())

        callables = self.key_listener._get_callables(
            pygame.Event(pygame.KEYDOWN, key=pygame.K_0, mod=pygame.KMOD_ALT)
        )

        callables2 = self.key_listener._get_callables(
            pygame.Event(pygame.KEYDOWN, key=pygame.K_1)
        )

        self.assertIn(test_func, callables.concurrent_functions)

        self.assertIn(test_func, callables2.concurrent_functions)

        bind2_dict = self.key_listener._key_hooks.get("test_bind2", {})
        bind2_event_dict = bind2_dict.get(True, {})
        bind2_list = bind2_event_dict.get(pygame.KEYDOWN, [])
        self.assertIn(test_func, bind2_list)

    def test_bind_sequential(self) -> None:

        @self.key_listener.sequential
        def test_func(_) -> None:
            pass

        self.key_listener.bind("test_bind0", pygame.K_0, pygame.KMOD_ALT)(test_func)

        self.key_listener.bind("test_bind1", pygame.K_1)(test_func)

        self.key_listener.bind("test_bind2")(test_func)

        self.assertIn("test_bind0", self.key_listener._key_hooks.keys())
        self.assertIn("test_bind1", self.key_listener._key_hooks.keys())
        self.assertIn("test_bind2", self.key_listener._key_hooks.keys())

        callables = self.key_listener._get_callables(
            pygame.Event(pygame.KEYDOWN, key=pygame.K_0, mod=pygame.KMOD_ALT)
        )

        callables2 = self.key_listener._get_callables(
            pygame.Event(pygame.KEYDOWN, key=pygame.K_1)
        )

        self.assertIn(test_func, callables.sequential_functions)

        self.assertIn(test_func, callables2.sequential_functions)

        bind2_dict = self.key_listener._key_hooks.get("test_bind2", {})
        bind2_event_dict = bind2_dict.get(False, {})
        bind2_list = bind2_event_dict.get(pygame.KEYDOWN, [])
        self.assertIn(test_func, bind2_list)

    def test_unbind(self) -> None:

        def test_func(_) -> None:
            pass

        self.key_listener.bind("test_bind0", pygame.K_0, pygame.KMOD_ALT)(test_func)

        self.key_listener.bind("test_bind1", pygame.K_1)(test_func)

        self.key_listener.bind("test_bind2")(test_func)

        self.key_listener.unbind(test_func, "test_bind0")

        bind0_dict = self.key_listener._key_hooks.get("test_bind0")
        assert bind0_dict
        bind0_list = bind0_dict.get(True, {}).get(pygame.KEYDOWN)
        self.assertNotIn(
            test_func,
            cast(list[Callable], bind0_list),
        )

        bind1_dict = self.key_listener._key_hooks.get("test_bind1")
        assert bind1_dict
        bind1_list = bind1_dict.get(True, {}).get(pygame.KEYDOWN)
        self.assertIn(
            test_func,
            cast(list[Callable], bind1_list),
        )

        bind2_dict = self.key_listener._key_hooks.get("test_bind2")
        assert bind2_dict
        bind2_list = bind2_dict.get(True, {}).get(pygame.KEYDOWN)
        self.assertIn(
            test_func,
            cast(list[Callable], bind2_list),
        )

    def test_clear_bind(self) -> None:

        def test_func(_) -> None:
            pass

        def test_func2(_) -> None:
            pass

        self.key_listener.bind("test_bind0", pygame.K_0, pygame.KMOD_ALT)(test_func)

        self.key_listener.bind("test_bind0")(test_func2)

        self.key_listener.bind("test_bind1", pygame.K_1)(test_func)

        self.key_listener.clear_bind("test_bind0")

        # No assertEmpty, so this will have to do.
        self.assertFalse(self.key_listener._key_hooks.get("test_bind0"))
        self.assertIsNotNone(self.key_listener._key_hooks.get("test_bind0"))
        bind1_dict = self.key_listener._key_hooks.get("test_bind1")
        assert bind1_dict
        bind1_list = bind1_dict.get(True, {}).get(pygame.KEYDOWN)
        self.assertIn(
            test_func,
            cast(list[Callable], bind1_list),
        )

        self.key_listener.clear_bind("test_bind0", True)

        self.assertIsNone(self.key_listener._key_hooks.get("test_bind0"))

    def test_bind_method(self) -> None:

        class TestClass:
            @self.key_listener.bind_method("test_bind")
            def test_method(self, _):
                pass

        self.assertHasAttr(TestClass.test_method, "_assigned_managers")
        assigned_listener_sets: list[
            tuple[KeyListener, str, int | None, int | None, int]
        ] = getattr(TestClass.test_method, "_assigned_managers")
        assigned_listeners = [set[0] for set in assigned_listener_sets]
        self.assertIn(self.key_listener, assigned_listeners)

    def test_register_class(self) -> None:

        @self.key_listener.register_class
        class TestClass:
            @self.key_listener.bind_method("test_bind")
            def test_method(self, _):
                pass

        test_instance = TestClass()

        # Verify attribute cleanup
        self.assertNotHasAttr(TestClass.test_method, "_assigned_listeners")
        # Verify class in assigned classes
        self.assertIn(TestClass, self.key_listener._assigned_classes.keys())
        # Verify method in listeners
        self.assertIn(
            TestClass.test_method, self.key_listener._class_listener_binds.keys()
        )
        self.assertIn(
            test_instance,
            cast(
                list[TestClass],
                self.key_listener._class_listener_instances.get(TestClass),
            ),
        )
        bound_listeners: dict[int, list[tuple[Callable, Type[object]]]] = (
            self.key_listener._class_listeners.get("test_bind", {}).get(True, {})
        )
        self.assertTrue(bound_listeners)
        listeners: list[tuple[Callable, Type[object]]] = bound_listeners.get(
            pygame.KEYDOWN, []
        )
        self.assertTrue(listeners)
        listener_pair = (TestClass.test_method, TestClass)
        # Verify method/object pair are associated with the event
        self.assertIn(listener_pair, listeners)

    def test_unbind_method(self) -> None:

        @self.key_listener.register_class
        class TestClass:
            @self.key_listener.bind_method("test_bind")
            def test_method(self, _):
                pass

            @self.key_listener.bind_method("test_bind")
            def test_method2(self, _):
                pass

        # Variable is for GC purposes
        # Don't want to bother with turning gc off
        test_instance = TestClass()  # noqa: F841

        self.key_listener.unbind_method(TestClass.test_method)
        self.assertNotIn(
            TestClass.test_method, self.key_listener._class_listener_binds.keys()
        )
        bound_listeners: dict[int, list[tuple[Callable, Type[object]]]] = (
            self.key_listener._class_listeners.get("test_bind", {}).get(True, {})
        )
        self.assertTrue(bound_listeners)
        listeners: list[tuple[Callable, Type[object]]] = bound_listeners.get(
            pygame.KEYDOWN, []
        )
        self.assertTrue(listeners)
        listener_pair = (TestClass.test_method, TestClass)
        # Verify method/object pair are associated with the event
        self.assertNotIn(listener_pair, listeners)

    def test_deregister_class(self) -> None:

        @self.key_listener.register_class
        class TestClass:
            @self.key_listener.bind_method("test_bind")
            def test_method(self, _):
                pass

        # Variable is for GC purposes
        # Don't want to bother with turning gc off
        test_instance = TestClass()  # noqa: F841

        self.key_listener.deregister_class(TestClass)
        # Verify method removed from listeners
        self.assertNotIn(
            TestClass.test_method, self.key_listener._class_listener_binds.keys()
        )
        self.assertNotIn(TestClass, self.key_listener._class_listener_instances.keys())
        bound_listeners: dict[int, list[tuple[Callable, Type[object]]]] = (
            self.key_listener._class_listeners.get("test_bind", {}).get(True, {})
        )
        self.assertTrue(bound_listeners)
        listeners: list[tuple[Callable, Type[object]]] = bound_listeners.get(
            pygame.KEYDOWN, []
        )
        listener_pair = (TestClass.test_method, TestClass)
        # Verify method/object pair are associated with the event
        self.assertNotIn(listener_pair, listeners)

    def test_notify_concurrent(self) -> None:

        example_var = False
        lock = threading.Lock()

        def test_func(_) -> None:
            nonlocal example_var
            lock.acquire()
            example_var = True
            lock.release()

        self.key_listener.bind("test_bind0", pygame.K_0, pygame.KMOD_ALT)(test_func)

        self.key_listener.bind("test_bind9", pygame.K_9, None)(test_func)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="1", key=pygame.K_1, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        # False, because the wrong key was pressed
        self.assertFalse(example_var)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="0", key=pygame.K_0, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        # False, because Alt isn't held
        self.assertFalse(example_var)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="0", key=pygame.K_0, mod=pygame.KMOD_ALT
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        # True, because both 0 and Alt are pressed
        self.assertTrue(example_var)

        # Reset example var for other binding test
        example_var = False

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="9", key=pygame.K_9, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        # True, because exact key combo match
        self.assertTrue(example_var)

        # Reset again
        example_var = False

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="9", key=pygame.K_9, mod=pygame.KMOD_ALT
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        # True, despite Alt also being pressed
        self.assertTrue(example_var)

    def test_notify_class_concurrent(self) -> None:

        lock = threading.Lock()

        @self.key_listener.register_class
        class TestClass:
            """
            Simple class for testing
            """

            def __init__(self):
                self.test_var = False

            @self.key_listener.bind_method("test_bind", pygame.K_0)
            def test_method(self, _):
                lock.acquire()
                self.test_var = True
                lock.release()

        test_class_list: list[TestClass] = []

        for i in range(3):
            test_class_list.append(TestClass())

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="9", key=pygame.K_9, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        for item in test_class_list:
            self.assertFalse(item.test_var)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="0", key=pygame.K_0, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        for item in test_class_list:
            self.assertTrue(item.test_var)

    def test_notify_sequential(self) -> None:

        example_var = False

        @self.key_listener.sequential
        def test_func(_) -> None:
            nonlocal example_var
            example_var = True

        self.key_listener.bind("test_bind0", pygame.K_0, pygame.KMOD_ALT)(test_func)

        self.key_listener.bind("test_bind9", pygame.K_9, None)(test_func)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="1", key=pygame.K_1, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_concurrent(event)
        # False, because the wrong key was pressed
        self.assertFalse(example_var)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="0", key=pygame.K_0, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_sequential(event)
        # False, because Alt isn't held
        self.assertFalse(example_var)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="0", key=pygame.K_0, mod=pygame.KMOD_ALT
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_sequential(event)
        # True, because both 0 and Alt are pressed
        self.assertTrue(example_var)

        # Reset example var for other binding test
        example_var = False

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="9", key=pygame.K_9, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_sequential(event)
        # True, because exact key combo match
        self.assertTrue(example_var)

        # Reset again
        example_var = False

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="9", key=pygame.K_9, mod=pygame.KMOD_ALT
        )
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.key_listener.notify_sequential(event)
        # True, despite Alt also being pressed
        self.assertTrue(example_var)

    def test_notify_class_sequential(self) -> None:

        @self.key_listener.register_class
        class TestClass:
            """
            Simple class for testing
            """

            def __init__(self):
                self.test_var = False

            @self.key_listener.bind_method("test_bind", pygame.K_0)
            @self.key_listener.sequential
            def test_method(self, _):
                self.test_var = True

        test_class_list: list[TestClass] = []

        for i in range(3):
            test_class_list.append(TestClass())

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="9", key=pygame.K_9, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.key_listener.notify_sequential(event)
        for item in test_class_list:
            self.assertFalse(item.test_var)

        local_event = pygame.event.Event(
            pygame.KEYDOWN, unicode="0", key=pygame.K_0, mod=pygame.KMOD_NONE
        )
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.key_listener.notify_sequential(event)
        for item in test_class_list:
            self.assertTrue(item.test_var)


if __name__ == "__main__":
    pygame.init()
    unittest.main()



from __future__ import annotations

from abc import ABC, abstractmethod
import json
import pathlib
from typing import TextIO, Type

from .key_map import KeyMap, KeyBind
from .joy_map import JoyMap

import pygame


def _get_parser_from_path(path: pathlib.Path) -> Type[FileParser]:
    file_type = path.suffix
    parser = None
    match file_type:
        case ".json":
            parser = JSONParser
        case _:
            raise ValueError(
                f'File type "{file_type}" is not implicitly supported. Please '
                "explicitly name a FileParser type."
            )
    return parser


class FileParser(ABC):

    @staticmethod
    @abstractmethod
    def load(in_file: TextIO) -> tuple[KeyMap, JoyMap]: ...

    @staticmethod
    @abstractmethod
    def save(key_map: KeyMap, joy_map: JoyMap, out_file: TextIO) -> None: ...

    @staticmethod
    @abstractmethod
    def _unpack_keys(maps: dict) -> dict: ...

    @staticmethod
    @abstractmethod
    def _unpack_joystick(maps: dict) -> dict: ...


class JSONParser(FileParser):

    @staticmethod
    def load(in_file: TextIO) -> tuple[KeyMap, JoyMap]:
        """
        Converts the given JSON file into a KeyMap and JoyMap

        :param in_file: Target file with the required data.
        :return: Created KeyMap and JoyMap
        """
        maps: dict = json.load(in_file)
        key_map = KeyMap()
        key_map.key_binds = JSONParser._unpack_keys(maps.get("keys", {}))
        joy_map = JoyMap()
        joy_map._joy_binds = JSONParser._unpack_joystick(maps.get("controller", {}))
        return key_map, joy_map

    @staticmethod
    def save(key_map: KeyMap, joy_map: JoyMap, out_file: TextIO) -> None:
        """
        Saves the KeyMap and JoyMap as a JSON string

        :param key_map: KeyMap object being saved
        :param out_file: File receiving the data
        """
        bind_keys = key_map.pack_binds()
        bind_joysticks = joy_map.pack_binds()
        maps = {"keys": bind_keys, "controller": bind_joysticks}
        json.dump(maps, out_file, indent=2)

    @staticmethod
    def _unpack_keys(maps: dict) -> dict:
        """
        Converts the JSON-styled dict into a dict compatible with a KeyMap

        :param maps: JSON-style dictionary of keybinds
        :return: Dictionary compatible with KeyMap
        """
        unpacked_dict: dict[int | None, list[KeyBind]] = {}
        for bind_name, key_data in maps.items():
            key_name, mod = key_data
            key_code = None
            if key_name is not None:
                key_code = pygame.key.key_code(key_name)
            unpacked_dict.setdefault(key_code, []).append(KeyBind(bind_name, mod))
            # binds = [KeyBind(bind_name=bind[0], mod=bind[1]) for bind in bind_list]
            # unpacked_dict.update({key_code: binds})
        return unpacked_dict

    @staticmethod
    def _unpack_joystick(maps: dict) -> dict:
        """
        Converts the JSON-styled dict into a dict compatible with a JoyMap

        :param maps: JSON-style dictionary of joystick binds
        :return: Dictionary compatible with JoyMap
        """
        unpacked_dict: dict[tuple, list[str]] = {}
        for bind_name, joy_data in maps.items():
            joy_data_points: list[tuple] = []
            for data_point in joy_data:
                joy_data_points.append((data_point[0], data_point[1]))
            fixed_joy_data: tuple = tuple(joy_data_points)
            unpacked_dict.setdefault(fixed_joy_data, []).append(bind_name)

        return unpacked_dict



import pathlib
import sys
import threading
from typing import Callable, cast, Type
import unittest
from weakref import WeakSet

import pygame

sys.path.append(str(pathlib.Path.cwd()))

from src.simple_events import getEventManager  # noqa: E402
from src.simple_events.event_manager import EventManager  # noqa: E402


class TestEventManager(unittest.TestCase):

    def assertHasAttr(self, obj, intendedAttr: str):
        testBool = hasattr(obj, intendedAttr)

        self.assertTrue(testBool, msg=f"{obj=} lacks an attribute, {intendedAttr=}")

    def assertNotHasAttr(self, obj, intendedAttr: str):
        testBool = hasattr(obj, intendedAttr)

        self.assertFalse(
            testBool, msg=f"{obj=} has unexpected attribute, {intendedAttr=}"
        )

    def setUp(self) -> None:
        self.event_manager = getEventManager("TestCase")
        self.test_event = pygame.USEREVENT + 1
        self.test_event2 = pygame.USEREVENT + 2

    def tearDown(self) -> None:
        self.event_manager._listeners.clear()

    def test_sequential_tag(self) -> None:

        @self.event_manager.register(self.test_event)
        @self.event_manager.sequential
        def test_func() -> None:
            pass

        self.assertHasAttr(test_func, "_runs_sequential")

    def test_concurrent_tag(self) -> None:

        @self.event_manager.register(self.test_event)
        @self.event_manager.concurrent
        @self.event_manager.sequential
        def test_func() -> None:
            pass

        self.assertNotHasAttr(test_func, "_runs_sequential")

    def test_register(self) -> None:

        @self.event_manager.register(self.test_event)
        def test_func() -> None:
            pass

        event_dict = self.event_manager._listeners.get(self.test_event, {})
        listeners = event_dict.get(True)
        self.assertIn(
            test_func,
            cast(list[Callable], listeners),
        )

    def test_register_sequential(self) -> None:

        @self.event_manager.register(self.test_event)
        @self.event_manager.sequential
        def test_func() -> None:
            pass

        @self.event_manager.register(self.test_event)
        def test_func2() -> None:
            pass

        event_dict = self.event_manager._listeners.get(self.test_event, {})
        conc_listeners = event_dict.get(True)
        seq_listeners = event_dict.get(False)
        self.assertNotIn(
            test_func,
            cast(list[Callable], conc_listeners),
        )
        self.assertIn(
            test_func,
            cast(list[Callable], seq_listeners),
        )
        self.assertNotIn(
            test_func2,
            cast(list[Callable], seq_listeners),
        )
        self.assertIn(
            test_func2,
            cast(list[Callable], conc_listeners),
        )

    def test_deregister(self) -> None:

        @self.event_manager.register(self.test_event)
        @self.event_manager.register(self.test_event2)
        def test_func(_) -> None:
            pass

        self.event_manager.deregister(test_func, self.test_event)

        event_dict = self.event_manager._listeners.get(self.test_event)
        assert event_dict
        call_list = event_dict.get(True)
        assert call_list is not None

        self.assertNotIn(test_func, call_list)

        event_dict2 = self.event_manager._listeners.get(self.test_event2)
        assert event_dict2
        call_list2 = event_dict2.get(True)
        assert call_list2

        self.assertIn(test_func, call_list2)

    def test_deregister_all(self) -> None:

        @self.event_manager.register(self.test_event)
        @self.event_manager.register(self.test_event2)
        def test_func(_) -> None:
            pass

        self.event_manager.deregister(test_func)

        event_dict = self.event_manager._listeners.get(self.test_event)
        assert event_dict
        call_list = event_dict.get(True)
        assert call_list is not None

        self.assertNotIn(test_func, call_list)

        event_dict2 = self.event_manager._listeners.get(self.test_event2)
        assert event_dict2
        call_list2 = event_dict2.get(True)
        assert call_list2 is not None

        self.assertNotIn(test_func, call_list2)

    def test_register_method(self) -> None:

        class TestClass:
            @self.event_manager.register_method(self.test_event)
            def test_method(self, _):
                pass

        self.assertHasAttr(TestClass.test_method, "_assigned_managers")
        assigned_managers_pairs: list[tuple[EventManager, int]] = getattr(
            TestClass.test_method, "_assigned_managers"
        )
        assigned_managers = [pair[0] for pair in assigned_managers_pairs]
        self.assertIn(self.event_manager, assigned_managers)

    def test_register_class(self) -> None:

        @self.event_manager.register_class
        class TestClass:
            @self.event_manager.register_method(self.test_event)
            def test_method(self, _):
                pass

        test_instance = TestClass()

        # Verify attribute cleanup
        self.assertNotHasAttr(TestClass.test_method, "_assigned_managers")
        # Verify class in assigned classes
        self.assertIn(TestClass, self.event_manager._assigned_classes.keys())
        # Verify method in listeners
        self.assertIn(
            TestClass.test_method, self.event_manager._class_listener_events.keys()
        )
        self.assertIn(
            test_instance,
            cast(
                WeakSet[TestClass],
                self.event_manager._class_listener_instances.get(TestClass),
            ),
        )
        event_dict = self.event_manager._class_listeners.get(self.test_event, {})
        listeners = event_dict.get(True, [])
        self.assertIsNotNone(listeners)
        listener_pair = (TestClass.test_method, TestClass)
        # Verify method/object pair are associated with the event
        self.assertIn(
            listener_pair, cast(list[tuple[Callable, Type[object]]], listeners)
        )

    def test_deregister_method(self) -> None:

        @self.event_manager.register_class
        class TestClass:
            @self.event_manager.register_method(self.test_event)
            def test_method(self, _):
                pass

        TestClass()

        self.event_manager.deregister_method(TestClass.test_method)

        # Verify method removed from listeners
        self.assertNotIn(
            TestClass.test_method, self.event_manager._class_listener_events.keys()
        )

        event_dict = self.event_manager._class_listeners.get(self.test_event, {})
        listeners = event_dict.get(True, [])
        self.assertIsNotNone(listeners)

        listener_pair = (TestClass.test_method, TestClass)
        # Verify method/object pair are detached from the event
        self.assertNotIn(
            listener_pair, cast(list[tuple[Callable, Type[object]]], listeners)
        )

    def test_deregister_class(self) -> None:

        @self.event_manager.register_class
        class TestClass:
            @self.event_manager.register_method(self.test_event)
            def test_method(self, _):
                pass

        TestClass()

        self.event_manager.deregister_class(TestClass)

        self.assertNotIn(TestClass, self.event_manager._assigned_classes.keys())

        # Verify method removed from listeners
        self.assertNotIn(
            TestClass.test_method, self.event_manager._class_listener_events.keys()
        )
        self.assertNotIn(TestClass, self.event_manager._class_listener_instances.keys())

        event_dict = self.event_manager._class_listeners.get(self.test_event, {})
        listeners = event_dict.get(True, [])
        self.assertIsNotNone(listeners)

        listener_pair = (TestClass.test_method, TestClass)
        # Verify method/object pair are detached from the event
        self.assertNotIn(
            listener_pair, cast(list[tuple[Callable, Type[object]]], listeners)
        )

    def test_event_purge(self) -> None:

        @self.event_manager.register(self.test_event)
        @self.event_manager.register(self.test_event2)
        def test_func(_) -> None:
            pass

        @self.event_manager.register(self.test_event)
        @self.event_manager.register(self.test_event2)
        def test_func2(_) -> None:
            pass

        self.event_manager.purge_event(self.test_event)
        # Verify both have been cleared from test_event

        self.assertNotIn(
            self.test_event,
            self.event_manager._listeners,
        )
        # But both need to remain in test_event2

        event_dict2 = self.event_manager._listeners.get(self.test_event2, {})
        self.assertIn(
            test_func,
            cast(list[Callable], event_dict2.get(True)),
        )
        self.assertIn(
            test_func2,
            cast(list[Callable], event_dict2.get(True)),
        )

    def test_notify_concurrent(self) -> None:

        example_var = False
        example_var2 = False
        lock = threading.Lock()

        @self.event_manager.register(self.test_event)
        def test_func(_) -> None:
            nonlocal example_var
            lock.acquire()
            example_var = True
            lock.release()

        @self.event_manager.register(self.test_event)
        @self.event_manager.sequential
        def test_func2(_) -> None:
            nonlocal example_var2
            example_var2 = True

        local_event = pygame.event.Event(self.test_event2)
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.event_manager.notify_concurrent(event)
        # Make sure only the correct event is responded to
        self.assertFalse(example_var)
        self.assertFalse(example_var2)
        local_event = pygame.event.Event(self.test_event)
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.event_manager.notify_concurrent(event)
        self.assertTrue(example_var)
        self.assertFalse(example_var2)

    def test_notify_class_concurrent(self) -> None:

        lock = threading.Lock()

        @self.event_manager.register_class
        class TestClass:
            """
            Simple class for testing
            """

            def __init__(self):
                self.test_var = False
                self.test_var2 = False

            @self.event_manager.register_method(self.test_event2)
            def test_method(self, _):
                lock.acquire()
                self.test_var = True
                lock.release()

            @self.event_manager.register_method(self.test_event2)
            @self.event_manager.sequential
            def test_method2(self, _):
                self.test_var2 = True

        test_class_list: list[TestClass] = []

        for _ in range(3):
            test_class_list.append(TestClass())

        local_event = pygame.Event(self.test_event)
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.event_manager.notify_concurrent(event)
        for item in test_class_list:
            self.assertFalse(item.test_var)
            self.assertFalse(item.test_var2)

        local_event = pygame.Event(self.test_event2)
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.event_manager.notify_concurrent(event)
        for item in test_class_list:
            self.assertTrue(item.test_var)
            self.assertFalse(item.test_var2)

    def test_notify_sequential(self) -> None:

        example_var = False
        example_var2 = False
        lock = threading.Lock()

        @self.event_manager.register(self.test_event)
        def test_func(_) -> None:
            nonlocal example_var
            lock.acquire()
            example_var = True
            lock.release()

        @self.event_manager.register(self.test_event)
        @self.event_manager.sequential
        def test_func2(_) -> None:
            nonlocal example_var2
            example_var2 = True

        local_event = pygame.event.Event(self.test_event2)
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.event_manager.notify_sequential(event)
        # Make sure only the correct event is responded to
        self.assertFalse(example_var)
        self.assertFalse(example_var2)
        local_event = pygame.event.Event(self.test_event)
        pygame.event.post(local_event)
        for event in pygame.event.get():
            self.event_manager.notify_sequential(event)
        self.assertFalse(example_var)
        self.assertTrue(example_var2)

    def test_notify_class_sequential(self) -> None:

        lock = threading.Lock()

        @self.event_manager.register_class
        class TestClass:
            """
            Simple class for testing
            """

            def __init__(self):
                self.test_var = False
                self.test_var2 = False

            @self.event_manager.register_method(self.test_event2)
            def test_method(self, _):
                lock.acquire()
                self.test_var = True
                lock.release()

            @self.event_manager.register_method(self.test_event2)
            @self.event_manager.sequential
            def test_method2(self, _):
                self.test_var2 = True

        test_class_list: list[TestClass] = []

        for _ in range(3):
            test_class_list.append(TestClass())

        local_event = pygame.Event(self.test_event)
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.event_manager.notify_sequential(event)
        for item in test_class_list:
            self.assertFalse(item.test_var)
            self.assertFalse(item.test_var2)

        local_event = pygame.Event(self.test_event2)
        pygame.event.post(local_event)

        for event in pygame.event.get():
            self.event_manager.notify_sequential(event)
        for item in test_class_list:
            self.assertFalse(item.test_var)
            self.assertTrue(item.test_var2)


if __name__ == "__main__":
    pygame.init()
    unittest.main()



from .base_manager import managerBasicConfig  # noqa: F401
from .event_manager import getEventManager, notifyEventManagers  # noqa: F401, E501
from .key_manager import getKeyListener, notifyKeyListeners  # noqa: F401, E501
from .file_parser import JSONParser  # noqa: F401, E501


def basicConfig(*args, **kwds):
    managerBasicConfig(*args, **kwds)



from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from dataclasses import dataclass, field
import functools
import threading
from typing import Callable, Type
from weakref import WeakSet

import pygame


@dataclass
class _CallableSets:
    """
    A collection of callables, broken up by type
    """

    concurrent_functions: list[Callable] = field(default_factory=list)
    sequential_functions: list[Callable] = field(default_factory=list)
    concurrent_methods: list[tuple[Callable, Type[object]]] = field(
        default_factory=list
    )
    sequential_methods: list[tuple[Callable, Type[object]]] = field(
        default_factory=list
    )


class _BaseThreadSystem(ABC):

    @abstractmethod
    def start_thread(self, callable: Callable, *args) -> None: ...


class DefaultThreadSystem(_BaseThreadSystem):

    def start_thread(self, callable, *args):
        threading.Thread(target=callable, args=args).start()


class AsyncThreadSystem(_BaseThreadSystem):
    def start_thread(self, callable, *args):
        asyncio.create_task(callable(*args))


class BaseManager(ABC):
    thread_system: _BaseThreadSystem = DefaultThreadSystem()

    def __init__(self, handle: str) -> None:
        self.handle: str = handle
        # Registered object as key, instances of object as values
        self._class_listener_instances: dict[Type[object], WeakSet[object]] = {}
        # Assigned object as key, associated methods as values
        self._assigned_classes: dict[Type[object], list[Callable]] = {}

    # def __init_subclass__(cls) -> None:
    #     cls.thread_system = DefaultThreadSystem()

    def sequential(self, func: Callable) -> Callable:
        """
        Marks a function as to be run sequentially.

        :param func: The function to be tagged
        :return: the tagged function
        """
        # None is smallest data type at 16 bytes
        # Lowest impact from being attached long term.
        setattr(func, "_runs_sequential", None)
        return func

    def concurrent(self, func: Callable) -> Callable:
        """
        Marks a function as to be run concurrently via threading.
        Removes a sequential mark if one exists.

        :param func: The function to be cleared
        :return: the cleared function
        """
        if hasattr(func, "_runs_sequential"):
            delattr(func, "_runs_sequential")
        return func

    def register_class(self, cls: Type[object]) -> Type[object]:
        """
        Prepares a class for event handling.
        This will hijack the class's init method to push its instances to
        the registering body.

        It will also go through and clean up the assigned methods.
        """
        # Mypy will throw an error here because it thinks this is illegal.
        # Hijacking an init is illegal? Guess I'm going to jail then.
        cls.__init__ = self._modify_init(cls.__init__)  # type: ignore
        # Add all of the tagged methods to the callables list
        for method in cls.__dict__.values():
            if not hasattr(method, "_assigned_managers"):
                continue  # No point checking something untagged.
            _assigned_managers = getattr(method, "_assigned_managers", [])
            self._verify_manager(cls, method, _assigned_managers)
            if len(_assigned_managers) == 0:
                # We cleaned up the assignments to this handler, but other handlers
                # might have yet to check. If all have cleaned up, we can remove the
                # hanging attribute.
                delattr(method, "_assigned_managers")
                # Now there's no sign we modified the method.

        return cls

    def notify(self, event: pygame.Event) -> None:
        """
        Finds all listeners for a given event, and calls them in their respective modes

        :param event: Target event for the listeners
        """
        callables = self._get_callables(event)
        self._handle_concurrent(event, callables)
        self._handle_sequential(event, callables)

    def notify_concurrent(self, event: pygame.Event) -> None:
        """
        Finds all listeners that are called concurrently for the given event, and calls
        them in a new thread

        :param event: Target event for the listeners
        """
        callables = self._get_callables(event)
        self._handle_concurrent(event, callables)

    def notify_sequential(self, event: pygame.Event) -> None:
        """
        Finds all listeners that are called sequentially for the given event, and calls
        them one at a time

        :param event: Target event for the listeners
        """
        callables = self._get_callables(event)
        self._handle_sequential(event, callables)

    @abstractmethod
    def _get_callables(self, event: pygame.Event) -> _CallableSets: ...

    def _handle_concurrent(self, event: pygame.Event, callables: _CallableSets) -> None:
        for function in callables.concurrent_functions:
            self.thread_system.start_thread(function, event)
            # threading.Thread(target=function, args=(event,)).start()
        for method, cls in callables.concurrent_methods:
            instances = self._class_listener_instances.get(cls, WeakSet())
            for instance in instances:
                self.thread_system.start_thread(method, instance, event)
                # threading.Thread(target=method, args=(instance, event)).start()

    def _handle_sequential(self, event: pygame.Event, callables: _CallableSets) -> None:
        for function in callables.sequential_functions:
            function(event)
        for method, cls in callables.sequential_methods:
            instances = self._class_listener_instances.get(cls, WeakSet())
            for instance in instances:
                method(instance, event)

    def _add_instance(self, cls: Type[object], instance: object) -> None:
        """
        Adds the instance into the collection of instances

        :param cls: The class of the instance
        :param instance: The new instance being captured
        """
        self._class_listener_instances.setdefault(cls, WeakSet()).add(instance)

    @abstractmethod
    def _capture_method(
        self, cls: Type[object], method: Callable, tag_data: tuple
    ) -> None: ...

    def _tag_method(self, method: Callable, tag_data: tuple) -> Callable:
        """
        Applies a tag attribute to a method that is being registered.

        :param method: Method being registered
        :param tag_data: Pertinent data for the concrete manager
        :return: The tagged method
        """
        # Tagging a method with an attribute for later reading?
        # This reeks of "cleverness"
        # Hope it's not too clever for me to debug.

        # Although I suppose if you're reading this, it got published, which means
        # I got it to work properly.
        assigned_managers: list[tuple[BaseManager, tuple]] = []
        if hasattr(method, "_assigned_managers"):
            # Deja vu? This isn't the first assignment, so we need to pull the
            # previous ones first.
            assigned_managers = getattr(method, "_assigned_managers", [])
        assigned_managers.append((self, tag_data))
        setattr(method, "_assigned_managers", assigned_managers)
        return method

    def _verify_manager(
        self,
        cls: Type[object],
        method: Callable,
        managers: list[tuple[BaseManager, tuple]],
    ) -> None:
        """
        Checks the list of assigned managers for a method and captures it if it is
        assigned to the calling manager

        :param cls: Class of the object being processed
        :param method: Method of cls being registered
        :param managers: list of managers and tag data.
        """
        _indexes_to_remove: list[int] = []
        for index, (manager, tag_data) in enumerate(managers):
            # A manager could be assigned multiple times for multiple events.
            if manager is not self:
                continue
            self._capture_method(cls, method, tag_data)
            # Whoops, undefined behavior
            # managers.pop(index)
            _indexes_to_remove.append(index)
            break
        # Need to clean up the processed indices to we can remove the tag attribute
        # from the method.
        for index in reversed(_indexes_to_remove):
            managers.pop(index)

    def _modify_init(self, init: Callable) -> Callable:
        """
        Extracts the class and instance being generated, and puts them into a
        dict, so that the method can be called upon it

        :param init: The initializer function of a class being registered.
        :return: The modified init function
        """
        functools.wraps(init)  # Needs this

        def wrapper(*args, **kwds):
            # args[0] of a non-class, non-static method is the instance
            # This is called whenever the class is instantiated,
            # and the instance is extracted and can be stored
            instance = args[0]
            cls = instance.__class__
            # No need to check for the instance, each only calls this once
            self._add_instance(cls, instance)
            # self._class_listener_instances.setdefault(cls, WeakSet()).add(instance)
            return init(*args, **kwds)

        return wrapper


def managerBasicConfig(*args, **kwds) -> None:
    if kwds.get("is_async", False):
        BaseManager.thread_system = AsyncThreadSystem()
    else:
        BaseManager.thread_system = DefaultThreadSystem()
